#include <string.h>   // for memcpy

#include <avr/io.h>
#include <avr/interrupt.h>

#include "net/netstack.h"
#include "net/ip.h"
#include "rprintf.h"
#include "ir.h"

/* timer0 runs freely at F_CPU (16Mhz) / 1024 and 
   thus has a resultion of 64us 
*/

#define MAX_IR_EVENTS 128  // store at most this number if ir events
static volatile unsigned char events = 0; 
static volatile unsigned char done = 0; 
static volatile unsigned char event_buffer[MAX_IR_EVENTS];

void ir_init (void) {
  TCCR0 = (1 << CS02) | (1 << CS00);  // pre-scaler 1024 for timer0
   
  MCUCR = ( MCUCR | _BV(ISC00) ) & ~_BV(ISC01); // irq on any change
  GIFR = _BV(INTF0);  // clear external interrupt 0 flag
  GICR |= _BV(INT0);  // enable extermal interrupt 0
}


#define IR_TIMEOUT  10000  // timeout in us
#define IR_TIMEOUT_RELOAD (256 - IR_TIMEOUT / 64)

// interrupt on timer0 overflow
SIGNAL(SIG_OVERFLOW0)
{
  // disable timer
  TIMSK &= ~_BV(TOIE0);       // timer0 overflow interrupt disable

  // having this irq happening means, that 
  // there was a timeout in capturing an ir signal
  done = 1;
}

void ir_service(void) {

  // check if some signal is in buffer
  if((done)&&(events)) {
    unsigned char *net_buffer = netstackGetBuffer() + ETH_HEADER_LEN + 
      IP_HEADER_LEN + UDP_HEADER_LEN;

    // send signal
//    rprintf("IR Signal, len = %d (max = %d)\n\r", 
//	    events, MAX_IR_EVENTS);
    net_buffer[0] = DSP_CMD_IR;
    memcpy(net_buffer+1, (void*)event_buffer, events);
    udpSend(0xffffffff, DSP_PORT, events+1, net_buffer);

    // restart capturing
    done = 0;
    events = 0;
  }
}

// external interrupt 0
SIGNAL(SIG_INTERRUPT0) {

  // if the timer irq is disabled, we are currently not
  // capturing a signal -> start capturing now
  if(!(TIMSK & _BV(TOIE0))) {
    // only start capturing if event buffer is empty
    if(events == 0) {
      // start timer
      TCNT0 = IR_TIMEOUT_RELOAD; 
      TIFR = _BV(TOV0);          // clear overflow flag
      TIMSK |= _BV(TOIE0);       // timer0 overflow interrupt enable
    }
  } else {
    // ir event while capturing is in progress and still space in buffer
    if(events < MAX_IR_EVENTS-1) {
      event_buffer[events++] = TCNT0 - IR_TIMEOUT_RELOAD;
      TCNT0 = IR_TIMEOUT_RELOAD;
    }
  }
}
