/*
  ledmatrix driver
*/

#include "global.h"

#include <stdio.h>
#include <avr/io.h>
#include <inttypes.h>
#include <string.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>

#include "timer.h"
#include "uart.h"
#include "rprintf.h"
#include "net/netstack.h"

#include "network.h"

#include "default_img.h"
#include "ledmatrix.h"

#include "font/font.h"  // include 5*8 font

#include "ir.h"

#define HZ    50
#define TIMER_RELOAD  (F_CPU / 32 / (DSP_PIX_Y * DSP_W * HZ))

volatile unsigned short ticks = 0;
volatile unsigned short seconds = 0;

// video ram 
unsigned char vram[DSP_MEM];

// special character codes < 0x20
#define CHR_END    0x00
#define CHR_RED    0x01
#define STR_RED    "\001"
#define CHR_GREEN  0x02
#define STR_GREEN  "\002"
#define CHR_YELLOW 0x03
#define STR_YELLOW "\003"
#define CHR_HOME   0x04
#define STR_HOME   "\004"
#define CHR_POS_X  0x05
#define STR_POS_X  "\005"
#define CHR_POS_Y  0x06
#define STR_POS_Y  "\006"

#define FONT_CHR_GAP   1  // empty pixels between characters
#define FONT_SPACE_GAP 1  // additional empty pixels for empty characters

// externl counters to do synchronous copying
static volatile u08 y = 0, dsp = 0;
static volatile u08 snd_cnt = 0;

u08 draw_char(u08 col, u08 x, u08 y, unsigned char chr) {
  const unsigned char *d;
  u08 r, c, rd;
  u08 s, e;

  if((chr < 0x20)||(chr > 128))
    return 0;

  // determine first and last used row
  d = font[chr-0x20];
  for(s=0;(s<5)&&(pgm_read_byte(d+s) == 0); s++);
  for(e=4;(e>0)&&(pgm_read_byte(d+e) == 0); e--);

  // empty character?
  if(e <= s) {
    s = 0;
    e = FONT_SPACE_GAP;
  }

  // start at first used row
  d = font[chr-0x20] + s;

  // draw all used rows incl. GAP
  for(rd=0,r=s;r<=e+FONT_CHR_GAP;r++) {
    u16 f = (x+rd)/PIXELS_PER_BYTE + y * DSP_BPL;
    u08 shift = 2 * ((x+rd)%PIXELS_PER_BYTE);
    u08 b = pgm_read_byte(d++);
    if(r > e) b = 0;

    for(c=0;c<8;c++) {
      
      if(b & (0x80 >> c)) {
	vram[f] &=   ~( (u08)(~(col | 0x3f)) >> shift);
	vram[f] |=   (col & 0xc0) >> shift;
      } else
	vram[f] &= ~(0xc0 >> shift);

      f += DSP_BPL;
    }

    rd++;
  }
  return rd;
}

u08 chr_width(unsigned char chr) {
  const unsigned char *d;
  u08 s, e;

  if((chr < 0x20)||(chr > 128))
    return 0;

  // determine first and last used row
  d = font[chr-0x20];
  for(s=0;(s<5)&&(pgm_read_byte(d+s) == 0); s++);
  for(e=4;(e>0)&&(pgm_read_byte(d+e) == 0); e--);

  // empty character?
  if(e <= s) {
    s = 0;
    e = FONT_SPACE_GAP;
  }

  return e - s + 1 + FONT_CHR_GAP;
}

// get width of pixels used by string on display
u08 str_width(char *str) {
  u08 len=0;
  while(*str) len += chr_width(*str++);
  return len;
}

void draw_text(u08 col, u08 x, u08 y, char *msg) {
  while(*msg) {
    x += draw_char(col, x, y, *msg++);
  }
}

// character output routine for rprintf, parse some special
// characters
void draw_chr(u08 chr) {
  static u08 x = 0, y = 0;  
  static u08 col = DSP_RED;
  static u08 last_chr = 0;

  if(last_chr) {
    switch(last_chr) {
      case CHR_POS_X:
	x = chr;
	break;
      case CHR_POS_Y:
	y = chr;
	break;
    }
    last_chr = 0;
    return;
  }

  switch(chr) {
    case CHR_RED:  // 0x01
      col = DSP_RED;
      break;

    case CHR_GREEN:  // 0x02
      col = DSP_GREEN;
      break;

    case CHR_YELLOW:  // 0x03
      col = DSP_YELLOW;
      break;

    case CHR_HOME:   // 0x04
      x = 0; y = 0;
      break;

    case CHR_POS_X:   // 0x05
      last_chr = CHR_POS_X;
      break;

    case CHR_POS_Y:   // 0x06
      last_chr = CHR_POS_Y;
      break;

    case '\r':    // 0x0a
      x = 0;
      break;

    case '\n':    // 0x0d
      y += 8;
      break;

    default:
      x += draw_char(col, x, y, chr);
      break;
  }
}

// delay for a minimum of <us> microseconds
// the time resolution is dependent on the time the loop takes
// e.g. with 4Mhz and 5 cycles per loop, the resolution is 1.25 us
void delay_us(unsigned short time_us) {
  unsigned short delay_loops;
  register unsigned short i;
  
  delay_loops = (time_us+3)/5*CYCLES_PER_US; // +3 for rounding up (dirty)
  
  // one loop takes 5 cpu cycles
  for (i=0; i < delay_loops; i++) { };
}

void delay_ms(unsigned short ms) {
  ms = ms * ((DSP_PIX_Y * DSP_W * HZ) / 1000);
  ticks = 0;

  while(ticks < ms);
}

// this is set true by the irq on every vsync
static volatile u08 vsync;

void ledmatrix_put(u08 *data) {
  for(vsync=0;!vsync;);  // wait for first row of first display
  memcpy(vram, data, DSP_MEM);
}

void ledmatrix_beep(void) {
  snd_cnt = 0xc0;
}

// this irq runs at 4khz and is the main display routine 
SIGNAL(SIG_OUTPUT_COMPARE2) {
  TCNT2 = 0;

  u08 x;                   // Zhlvars fr die Datenbits (Zeilen,Spalten)

  static int m;
  static unsigned short sub_secs = 0;

  // tick 
  ticks++;

  // make funny sound ...
  if(snd_cnt) {
    // high pitched sound for 64/4000 = 62.5 sek
    if((snd_cnt & 0xc0) == 0x80)
      PORTD ^= _BV(3);

    // pause for 64/4000 = 62.5 sek
    if((snd_cnt & 0xc0) == 0x40)
      PORTD &= ~_BV(3);

    // low pitched sound for 64/4000 = 62.5 sek
    if((snd_cnt & 0xc0) == 0x00)
      if(ticks & 1)
	PORTD ^= _BV(3);

      snd_cnt--;
  } else
    PORTD &= ~_BV(3);     // snd output off

  sbi(PORTC, 3);              //Bright auf High

  if(sub_secs++ >= (DSP_PIX_Y * DSP_W * HZ)) {
    sub_secs = 0;
    seconds++;
  }

  // switch to display
  PORTC = (PORTC & 0xf8) | dsp;

  // reset counters
  if(y == 0) {
    PORTB |=  _BV(4); //Reset high
    PORTB &= ~_BV(4); //Reset lo
  }

  u08 *d0 = vram + DSP_BYTES_X * dsp + y * DSP_BYTES_X * DSP_W;
  u08 *d1 = d0 + DSP_OFFSET;

  // shift one line into one display
  for (x = 0; x<DSP_BYTES_X; x++) {
    u08 a = *d0++;
    u08 b = *d1++;

    PORTD = (PORTD & 0x0f) | (0xf0 & (((a & 0xc0) >> 2) | ((b & 0xc0))));
    sbi(PORTC,4); cbi(PORTC,4);

    PORTD = (PORTD & 0x0f) | (0xf0 & (((a & 0x30)) | ((b & 0x30) << 2)));
    sbi(PORTC,4); cbi(PORTC,4);

    PORTD = (PORTD & 0x0f) | (0xf0 & (((a & 0x0c) << 2) | ((b & 0x0c) << 4)));
    sbi(PORTC,4); cbi(PORTC,4);

    PORTD = (PORTD & 0x0f) | (0xf0 & (((a & 0x03) << 4) | ((b & 0x03) << 6)));
    sbi(PORTC,4); cbi(PORTC,4);
  }

  y++;
  if(y == DSP_PIX_Y) {
    y = 0;

    dsp++;
    if(dsp == DSP_W) {
      dsp = 0;  
      vsync = 1;
      m++;
    }
  }

  sbi(PORTC,3);
  sbi(PORTC,3);
  sbi(PORTC,3);

  cbi(PORTC,3);              //Bright auf Low
}

#define IP_Y 22

// called from ip layer to display ip address
void display_ip(u32 ip) {
  static u08 oldlen = 0;  // width of last ip address
  char str[32];
  u08 i,x,newlen;

  // build string of ip address
  sprintf(str, "%d.%d.%d.%d", 
	  ((unsigned char*)&ip)[3],
	  ((unsigned char*)&ip)[2],
	  ((unsigned char*)&ip)[1],
	  ((unsigned char*)&ip)[0]);
  newlen = str_width(str);

  // erase old message first by painting a bunch of spaces
  if(newlen < oldlen) {
    u08 spc_width = chr_width(' ');
    for(i=0;i<oldlen/spc_width;i++)
      draw_char(DSP_RED, (DSP_W*DSP_PIX_X-oldlen)/2+i*spc_width, IP_Y, ' ');
  }

  for(x=(DSP_W*DSP_PIX_X-newlen)/2,i=0;i<strlen(str);i++)
    x += draw_char(DSP_RED, x, IP_Y, str[i]);

  oldlen = newlen;
}

int main( void ) {
  u16 i;

  uartInit();
  uartSetBaudRate(115200);
  rprintfInit(uartSendByte);
  rprintf("\r\n ** LEDMATRIX **\r\n");

  ir_init();

  rprintf("size of display memory: %d bytes\r\n", DSP_MEM);

  // clear video mem
  for(i=0;i<DSP_MEM;i++)
    vram[i] = 0;

  // audio out
  DDRD |= _BV(3);

  // dear default image
  memcpy_P(vram, default_img, DSP_MEM);
  rprintfInit(draw_chr);
  rprintf(STR_GREEN STR_POS_X "%c" STR_POS_Y "%cV" VERSION, 60, 13);
  rprintfInit(uartSendByte);

  DDRC |= 0x1f;    // CLOCK, BRIGHT, SEL2-0
  DDRB |= _BV(4);  // RESET
  DDRD |= 0xf0;    // RED0-1/GREEN0-1

  // toggle on compare, clk/32
  TCCR2 = _BV(WGM21) | _BV(CS21) | _BV(CS20); //  | _BV(COM20);
  TIMSK |= _BV(OCIE2);   // interrupt on compare
  OCR2 = TIMER_RELOAD;
  TCNT2 = 0;                //Timer 2 Resetten

  // restart update
  sbi(PORTB, 4);              // Reset high
  cbi(PORTB, 4);              // Reset low

  // display always on
  cbi(PORTC, 3);              //Bright auf Low

  sei();

  network_init();

  while(1) {
    ir_service();
    
    // service local stuff
    serviceLocal();
    // service the network
    netstackService();
  }

  return 0;
}
