/* gifread.c - The high level GIF reading routines.  See writegif for the
   write side.  Also gifdecode.c for lower level GIF reading code. */

#include <stdio.h>
#include <stdlib.h>

#include "gifcodes.h"
#include "image.h"

static FILE *gif_file;
static image_t *image;
static unsigned char colormap[256];
static int gif_line;

/* Get next byte from file for decoder.
 * return -1 at end of file. */
int gif_get_byte(void) {
  return(fgetc(gif_file));
}

/* Output a line of gif. */
int gif_out_line(unsigned char *pixels, int linelen) {
  int i;
  unsigned char *dst = image->data + gif_line * (image->width/4);

  for(i=0;i<linelen;i++) {
    *dst |= colormap[*pixels++] << (2*(3-(i&3)));
    if((i & 3) == 3) dst++;
  }

  gif_line++;
  //  memcpy(image->pixels + y*image->width, pixels, linelen);
  return(0);
}


/* Create memory image based on gif file. 
 * Note this is based on a very old gif reader
 * that only handles the GIF87a version. 
 * This is the same that mgSaveGif creates at
 * least.  This version of gif was always
 * color mapped. */

image_t *gif_read(char *name) {
  int c;
  char type[7];
  int gif_colors = 0;
  unsigned char gif_cmap[256][3];
  int scan_width;
  struct gif_header gif;
  struct gif_image gim;
  
  image = NULL;
  gif_line = 0;

  gif_file = fopen(name, "rb");
  if (fread(&gif, 1, sizeof(gif), gif_file) < sizeof(gif)) {
    goto TRUNCOUT;
  }

  memcpy(type, gif.giftype, 6);
  type[6] = 0;
  if (memcmp("GIF", type, 3) != 0) {  // (!startsWith("GIF", type)) 
    fprintf(stderr, "Not a good GIF file\n");
    goto BADOUT;
  }

#if 0
  if (strcmp("GIF87a", type) != 0) { // (!sameString("GIF87a", type)) 
    fprintf(stderr, "GIF: found version %s, sadly load_gif only speaks "
	     "version GIF87a\n", type);
    //    goto BADOUT;
  }
#endif

  gif_colors = (1<<((gif.colpix&PIXMASK)+1));
  if(gif.colpix&COLTAB) {
    printf("GIF: reading global colormap with %d colors\n", gif_colors);
    int size = gif_colors*3;
    if (fread(gif_cmap, 1, size, gif_file) < size)
      goto TRUNCOUT;
  }

  /* skip over extension blocks and other junk til get ',' */
  for (;;) {
    if ((c = fgetc(gif_file)) == READ_ERROR)
      goto TRUNCOUT;
    if (c == ',')
      break;
    if (c == ';')    /* semi-colon is end of piccie */
      goto TRUNCOUT;
    if (c == '!') {    /* extension block */
      if ((c = fgetc(gif_file)) == READ_ERROR)    /* skip extension type */
	goto TRUNCOUT;
      for (;;) {
	if ((c = fgetc(gif_file)) == READ_ERROR)
	  goto TRUNCOUT;
	if (c == 0)    /* zero 'count' means end of extension */
	  break;
	while (--c >= 0) {
	  if (fgetc(gif_file) == READ_ERROR)
	    goto TRUNCOUT;
	}
      }
    }
  }
  if (fread(&gim, 1, sizeof(gim), gif_file) < sizeof(gim))
    goto TRUNCOUT;

  image = malloc(sizeof(image_t));
  scan_width = image->width = (gim.whi<<8) + gim.wlo;
  // increase image width to multiple of 4
  if(image->width & 3) {
    printf("GIF: adjusting image width from %d to %d\n", 
	   image->width, (image->width+3) & ~3);
    image->width = (image->width+3) & ~3;
  }

  image->height = (gim.hhi<<8) + gim.hlo;  
  if(image->height != 32) {
    printf("GIF warning: unsupported image height!\n");
  }

  image->data = malloc(image->width * image->height / 4);
  memset(image->data, 0, image->width * image->height / 4);

  printf("GIF: image size %d x %d\n", image->width, image->height);

  /* Gif files can have color maps in two places.  Let
   * the gim color map overwrite the one in the gif header
   * here. */
  if (gim.flags&COLTAB) {
    int size;
    gif_colors = (1<<((gim.flags&PIXMASK)+1));
    printf("GIF: reading local colormap with %d colors\n", gif_colors);
    size = gif_colors*3;
    if (fread(gif_cmap, 1, size, gif_file) < size)
      goto TRUNCOUT;
  }

  if (gif_colors > 0) {
    int i, oor=0;

    if (gif_colors > 256)
      fprintf(stderr, "Too many colors in %s\n", name);

    printf("GIF: mapping colors\n");

    // map all colors to the displays four colors
    for(i=0;i<gif_colors;i++) {
      if(      (gif_cmap[i][0] < 10) && (gif_cmap[i][1] < 10) && 
	       (gif_cmap[i][2] < 10))
	colormap[i] = 0;
      else if((gif_cmap[i][0] > 245) && (gif_cmap[i][1] < 10) && 
	      (gif_cmap[i][2] < 10))
	colormap[i] = 1;
      else if((gif_cmap[i][0] < 10) && (gif_cmap[i][1] > 245) && 
	      (gif_cmap[i][2] < 10))
	colormap[i] = 2;
      else if((gif_cmap[i][0] > 245) && (gif_cmap[i][1] > 245) && 
	      (gif_cmap[i][2] < 10))
	colormap[i] = 3;
      else {
	oor++;
	colormap[i] = 0;
      }
    }
    if(oor != 0) 
      printf("GIF: %d of %d colors out of range\n", oor, gif_colors);
  }
  
  switch (gif_decoder(scan_width)) {
  case READ_ERROR:
  case BAD_CODE_SIZE:
    fprintf(stderr, "decoder failed\n");
    goto TRUNCOUT;
  case OUT_OF_MEMORY:
    fprintf(stderr, "out of memory\n");
    goto BADOUT;
  default:
    printf("GIF: decoder succeeded\n");
    break;
  }

  fclose(gif_file);
  return(image);
  
 TRUNCOUT:
  fprintf(stderr, "%s is truncated\n", name);
 BADOUT:
  fclose(gif_file);
  
  if(image != NULL) {
    free(image->data);
    free(image);
  }

  return(NULL);
}
